#!/bin/sh

rootmnt="/sysroot"

for x in $(cat /proc/cmdline); do
    if [ `echo $x|grep "root="` ]; then
        ROOT=${x#root=}
    fi
    if [ `echo $x|grep "boot="` ]; then
        BOOT=${x#boot=}
    fi

    # specify the writable partition
    if [ `echo $x | grep "overlayup="` ]; then
        WPDEV=${x#overlayup=}
    fi
    # specify the backup point
    if [ `echo $x | grep "overlayx="` ]; then
        BACKUPPOINT=${x#overlayx=}
    fi
done

echo "********** RootOverlay Begin! ***********"
echo "ROOT= ${ROOT} "
echo "BOOT= ${BOOT} "
echo "WPDEV= ${WPDEV}"
echo "BACKUPPOINT= ${BACKUPPOINT}"

function sort_by_index
{
    arr=$1
    prefix=$2

#echo ${arr[@]}

    declare -a numarr
    declare -a res

    prefixlen=${#prefix}

    for el in ${arr[*]};do
        if [[ ! ${el:0:$prefixlen} == $prefix || ! ${el:$prefixlen} -gt 0 ]];then
            continue
        fi

        num=${el:$prefixlen}
        numarr=("${numarr[@]}" $num)
    done

#echo ${numarr[@]}

   sortedstr=$(echo ${numarr[@]} | tr ' ' '\n' | sort -n)
   for el in $sortedstr;do
       res=("${res[@]}" $prefix$el)
   done

   echo "${res[@]}"
}

process_restore()
{
	if [[ -e /upper/.restore.flag ]];then
		echo "process restore,delete upper_now and upper_boot_now"

		tormdir=/upper/upper_now
		for file in `ls $tormdir 2>/dev/null`;do
		        if [[ $file == "etc" ]];then
                		echo rm -rf "${tormdir}/etc/*"
               		rm -rf "${tormdir}/etc/*"
		        else
		                echo rm -rf "${tormdir}/$file"
		                rm -rf "${tormdir}/$file"
		        fi
		done

		#rm -rf /upper/upper_now/*
		rm -rf /upper/upper_boot_now/*

		rm -f /upper/.restore.flag
	fi
}

process_deleted_bakcups()
{
	tmpbak_upper_dir=/upper/.tmpbakupper/
	tmpbak_upper_boot_dir=/upper/.tmpbakupperboot/
	
	upperdirs=`ls /upper/ 2>/dev/null`
	sorted_upperdirs=$(sort_by_index "${upperdirs[*]}" upper)
	for file in $sorted_upperdirs; do 
		index=${file:5}
		if [[ ! $index -gt 0 ]] 2>/dev/null;then 
			continue
		fi 
		
		mkdir -p $tmpbak_upper_dir
		mkdir -p $tmpbak_upper_boot_dir
	
		if [[ -e /upper/.delete.flag.$index ]];then 
			echo "rm backup $index"

			cp -aR /upper/${file}/* $tmpbak_upper_dir 2>/dev/null
			cp -aR /upper/upper_boot${index}/* $tmpbak_upper_boot_dir 2>/dev/null
			rm -rf /upper/${file}
			rm -rf /upper/upper_boot${index}
			
			rm -rf /upper/.delete.flag.$index
			rm -rf /upper/desc$index
			rm -rf /upper/name$index
			rm -rf /upper/time$index
			continue
		fi 

		# 以下为不删除的备份，要考虑合入在前面的已经被标记为删除的备份，因为依赖到他们的文件
		if [[ ! "`ls -A $tmpbak_upper_dir 2>/dev/null`" == "" ]];then 
			cp -aR /upper/${file}/* $tmpbak_upper_dir
			rm -rf /upper/${file}
			mv $tmpbak_upper_dir /upper/${file}
			# support selinux
			echo "" > /upper/.needrelabel.flag.$index
		fi 
		if [[ ! "`ls -A $tmpbak_upper_boot_dir 2>/dev/null`" == "" ]];then 
			cp -aR /upper/upper_boot${index}/* $tmpbak_upper_boot_dir
			rm -rf /upper/upper_boot${index}
			mv $tmpbak_upper_boot_dir /upper/upper_boot${index}
			# support selinux
			echo "" > /upper/.needrelabel.flag.$index
		fi 
	done
	
	return 0
}

overlay_root()
{
    if [[ ${WPDEV} == "" ]]; then
        WPDEV=/dev/sdb
	echo "set WPDEV: $WPDEV"
    fi

    umount ${rootmnt}
    if [ $? -ne 0 ];then
	echo "umount $rootmnt failed"
	return 1
    fi
    mkdir -p ${rootmnt}.real
    mount $ROOT ${rootmnt}.real
    if [ $? -ne 0 ];then
        echo "mount $ROOT to ${rootmnt}.real failed"
        mount $ROOT $rootmnt
	return 1
    fi

    mkdir -p /upper
    mount ${WPDEV} /upper
    if [[ $? -ne 0 ]]; then
        echo "[do_overlay] mount $WPDEV to /upper failed"
        umount ${rootmnt}.real
        mount $ROOT $rootmnt
        return 1
    fi

    # need relabel first time, for example: after installing os
    if [[ ! -e /upper/upper_now || ! -e /upper/upper_boot_now || ! -e /upper/work || ! -e /upper/work_boot ]];then
        echo "" > /upper/.needrelabel.flag.-1
    fi

	
	process_deleted_bakcups
	if [[ $? -ne 0 ]];then
		echo "[do_overlay] process deleted backups failed"
		umount ${rootmnt}.real
		mount $ROOT $rootmnt
		return 1
	fi

	process_restore
	if [[ $? -ne 0 ]];then
		echo "[do_overlay] process restore failed"
		umount ${rootmnt}.real
		mount $ROOT $rootmnt
		return 1
	fi

    mkdir -p /upper/work
    mkdir -p /upper/upper_now

    lowerdirs=${rootmnt}.real
    upperdirs=`ls /upper/ 2>/dev/null` 
	sorted_upperdirs=$(sort_by_index "${upperdirs[*]}" upper)
    for dir in $sorted_upperdirs; do
        if [[ ! ${dir:0:5} == "upper" || $dir == "upper_now" || $dir == "upper" ]];then
			continue
        fi
        index=${dir:5}
        if [[ -z "$BACKUPPOINT" || "$BACKUPPOINT" -ge $index ]];then
            lowerdirs=/upper/${dir}:${lowerdirs}
        fi
    done

    mount -t overlay overlay -olowerdir=${lowerdirs},upperdir=/upper/upper_now,workdir=/upper/work ${rootmnt}
    if [[ $? -ne 0 ]]; then
	echo "[do_overlay] mount /lower and /upper to /sysroot failed"
	umount /upper
	umount ${rootmnt}.real
        mount $ROOT $rootmnt
	return 1
    fi

    mkdir -p ${rootmnt}/.overlayup
    mount --bind /upper ${rootmnt}/.overlayup
    if [[ $? -ne 0 ]]; then
	echo "[do_overlay] mount --bind /upper ${rootmnt}/.overlayup failed"
	umount ${rootmnt}
	umount /upper
	umount ${rootmnt}.real
        mount $ROOT $rootmnt
	return 1
    fi

    echo "make root overlay succeed"
    return 0
}

overlay_boot()
{
    if [[ -z "$BOOT" ]];then
	echo "can not get boot partition, do not overlay boot"
	return 0
    fi

    mkdir -p ${rootmnt}/boot
    mkdir -p ${rootmnt}/.boot.real
    mount ${BOOT} ${rootmnt}/.boot.real
    if [[ $? -ne 0 ]];then
        echo "mount $BOOT} ${rootmnt}/boot.real failed"
        return 1
    fi
    mkdir -p /upper/upper_boot_now
    mkdir -p /upper/work_boot

    lowerdirs=${rootmnt}/.boot.real
    upperdirs=`ls /upper/ 2>/dev/null` 
	sorted_upperdirs=$(sort_by_index "${upperdirs[*]}" upper_boot)
    for dir in $sorted_upperdirs; do
        if [[ ! ${dir:0:10} == "upper_boot" || $dir == "upper_boot_now" ]];then
			continue
        fi
        index=${dir:10}
        if [[ -z "$BACKUPPOINT" || "$BACKUPPOINT" -ge $index ]];then
            lowerdirs=/upper/${dir}:${lowerdirs}
        fi
    done

    mount -t overlay overlay -olowerdir=${lowerdirs},upperdir=/upper/upper_boot_now,workdir=/upper/work_boot ${rootmnt}/boot
    if [[ $? -ne 0 ]]; then
	echo "mount /lower and /upper to ${rootmnt}/boot failed"
	umount ${rootmnt}/.boot.real
	return 1
    fi

    # delete /boot item, if /etc/fstab has been changed,do not change it again
    cat /etc/fstab | grep " /boot " > /dev/null
    if [[ $? -eq 0 ]];then
        cat ${rootmnt}.real/etc/fstab | grep -v " /boot " > /fstab.tmp
        echo "# this file is changed by rootoverlay" >> /fstab.tmp
        mv /fstab.tmp ${rootmnt}/etc/fstab
    fi

    echo "make boot overlay succeed !"
    return 0
}

do_overlay() 
{
    cat /proc/cmdline | grep "overlayup" > /dev/null
    if [ $? -ne 0 ];then
        echo "rootoverlay disabled, exit"
        return
    fi

    overlay_root
    if [[ $? -ne 0 ]];then
        echo "overlay_root failed"
        return
    fi

    overlay_boot
    if [[ $? -ne 0 ]];then
	echo "overlay_boot failed"
        umount ${rootmnt}
        umount /upper
        umount ${rootmnt}.real
        mount $ROOT $rootmnt    
	return
    fi

}

do_overlay

echo "********** RootOverlay Finished ! ***********"

